#CASA script for CASA-VLBI paper
#
# This is a transcript from the default N14C3 notebook
# Dependencies: casavlbitools, natsort
# Verified on CASA6.5.1
#
# IVB 220916


# Global imports
from time import time
import glob
import shutil
import os
from natsort import natsort_keygen
from casavlbitools import fitsidi


##################################
# Settings for the user to change
##################################

# Experiment settings
mypath="set your path here"
outpath=mypath
datapath=mypath
obsid="n14c3"

# Stamp for reruns
stamp = 'CASA'

# Calibration settings
# Reference antenna
refant = ''
# Single band delay
sbdscan= ''
sbdtimer = ''
# Multi-band delay
mbdfield = ''
# Bandpass
bpassfield = ''

# Hardcode edge channels to flag
edgechannels=''

# Imaging settings
# As imaging is a seperate lecture, these settings are pre-filled)
target1 = 'J1849+3024'
phasecal1 = '1848+283'
target2 = '3C345'
phasecal2 = 'J1640+3946'

allsources=[target1,phasecal1,target2,phasecal2]

cell=['0.3mas']
imsize = [512,512]
niter = 500

# NO CHANGES NEEDED BELOW THIS POINT

##################################
# Derived parameters
##################################

#Visibility data (needed for pre-calibration steps)
# In the case where you have more than 9 FITS-IDI files, you need to
# sort them with natural sorting, using natsort
#idifiles = sorted(glob.glob(datapath+'*_1_1.IDI*'),key=natsort_keygen())
idifiles = sorted(glob.glob(datapath+'*_1_1.IDI*'))
vis = mypath+obsid+'.ms'

#EVN archival calibration tables (needed for pre-calibration steps)
AIPSflag = datapath+obsid+'.uvflg'
antabfile = datapath+obsid+'.antab'

#CASA calibration tables
gcaltab = mypath+obsid+'_'+stamp+'.gcal'
tsystab = mypath+obsid+'_'+stamp+'.tsys'
sbdtab = mypath+obsid+'_'+stamp+'.sbd'
mbdtab = mypath+obsid+'_'+stamp+'.mbd'
bpasstab = mypath+obsid+'_'+stamp+'.bpass'

flagfile = mypath+obsid+'.flag'
myflags = mypath+obsid+'.myflags'
listobsfile = mypath+obsid+'.listobs'

#########################
# Pre-calibration steps
#########################
# Step 1: Append Tsys
try:
    print('Appending system temperature information, this can take some time.')
    fitsidi.append_tsys(antabfile, idifiles)
except RuntimeError:
    print('\U00002705 System temperature already present, continuing...')
except FileNotFoundError:
    print('\U0001F6D1 Your FITS-IDI files cannot be found, have you set the correct path?')
except IndexError:
    print("\U0001F6D1 Your list of files is empty, have you set the correct path?")
    
# Step 2: Append the gain curve (new method!)
try:
    print('Appending gain information.')
    fitsidi.append_gc(antabfile,idifiles[0])
except RuntimeError:
    print('\U00002705 Gain curve already present. Moving on!')
    
# Step 3: Convert UVFLG
fitsidi.convert_flags(AIPSflag,idifiles,outfile=flagfile)

# Step 4: Import data to MS format
#Verify if MS exist, otherwise do importstep
if os.path.isdir(vis)==False:
    print('Starting the import to MS, this takes time, be patient')
    importfitsidi(fitsidifile= idifiles, vis = vis, constobsid=True, scanreindexgap_s=15.0, specframe='GEO')
else:
    print("Import already done, clearing calibration and flags")
    clearcal(vis)
    flagmanager(vis=vis,mode='restore',versionname='archival_flags')

# Step 5: Apply flags from archive
# For some reason this now prints a parameter list, but works as it should
flagdata(vis=vis, mode='list', inpfile=flagfile,
         reason='any', action='apply', flagbackup=False, savepars=False)
flagmanager(vis=vis,mode='save',versionname='archival_flags',
        comment='Archival flags prior to any calibration')

# Step 6: Generate list of scans
if os.path.isfile(listobsfile)==False:
    listobs(vis, listfile=listobsfile)
else:
    print("\U00002705 A file with listobs information is already present")

##########################################
# Generating TY and GC calibration tables
##########################################
# gencal generates a cumulative table if the table exists, remove old tables first
if os.path.isdir(tsystab)==True:
    shutil.rmtree(tsystab)

gencal(vis,caltable=tsystab,caltype='tsys',uniform=False)

if os.path.isdir(gcaltab)==True:
    shutil.rmtree(gcaltab)

gencal(vis,caltable=gcaltab,caltype='gc')

#applycal(vis=vis,gaintable=[tsystab,gcaltab], flagbackup=False, parang=True)

#plotms(vis=vis, xaxis='frequency', yaxis='amp',ydatacolumn='data', 
#    antenna=refant, correlation='ll',coloraxis='baseline', 
#    scan=sbdscan,averagedata=True, avgtime='600')

##########################################
# Flagging
##########################################
# Autocorrelations
flagdata(vis,mode='manual',autocorr=True,flagbackup=False)

# Bad data flags from myflags file
if os.path.isfile(myflags):
    flagdata(vis, mode='list', inpfile=myflags, flagbackup=False)

# Bad stations
flagdata(vis, mode='manual',antenna='SV', flagbackup=False)

# Flag the edge channels
# Determine the number of spectral windows and channels in the MS
# Needed for edge-channel flagging and for spectral window mapping 
msmd.open(vis)
nspw=msmd.nspw()
nchan=msmd.nchan(1)
msmd.done()
# By default flag 10% of the edge channels
edgefraction=0.1
try:
    flagfraction = int(nchan/(100*edgefraction)) 
    start = str(flagfraction-1)
    end = str(nchan-flagfraction)
    spwflag = '*:0~'+start+';'+end+'~'+str(nchan-1)
    flagdata(vis,mode='manual',spw=spwflag,flagbackup=False)
except ZeroDivisionError:
    print("Not flagging the edge channels")
    
# Quack the first 5 seconds
flagdata(vis,mode='quack',quackinterval=5,flagbackup=False)

# Write flagversion entry
flagmanager(vis,mode='save',versionname='precal_flags',
    comment='Flags from Tsys, gaincal, bad data and edge channels')

#####################################
# Fringe fitting
#####################################
# Single band delay
fringefit(vis, caltable=sbdtab, timerange=sbdtimer, 
        solint='inf', zerorates=True, refant=refant,minsnr=50,
        gaintable=[gcaltab,tsystab],
        interp=['nearest','nearest,nearest'],
        parang=True)


# Multi band delay
fringefit(vis, caltable=mbdtab, field=mbdfield,
        solint='inf', zerorates=False, refant=refant,
        combine='spw', minsnr=5,
        gaintable=[gcaltab, tsystab, sbdtab],
        interp=['nearest','nearest,nearest','nearest'],
        parang=True)

#####################################
# Bandpass
#####################################
bandpass(vis, caltable=bpasstab, field=bpassfield,
        gaintable=[gcaltab, tsystab, sbdtab, mbdtab],
        interp=['nearest','nearest,nearest','nearest','linear'],
        spwmap=[[],[],[], int(nspw)*[0]],
        solnorm=True, solint='inf', refant=refant, 
        bandtype='B', parang=True)

#####################################
# Apply calibration and split data
#####################################
applycal(vis,
    gaintable=[gcaltab, tsystab, sbdtab, mbdtab, bpasstab],
    interp=['nearest','nearest,nearest','nearest','linear','linear,linear'],
    spwmap=[[], [], [], int(nspw)*[0],[]],
    parang=True)


#####################################
# Images
#####################################
# From here the data is ready to be imaged in CASA task tclean
# Example run:

tclean(vis=vis, field=target1, imagename='clean_'+target1,imsize=imsize, cell=cell,niter=niter)
imview(raster='clean_'+target1+'.image')
